@@ -43,6 +43,7 @@ void orc_compiler_check_sizes (OrcCompiler *compiler);
 
 static char **_orc_compiler_flag_list;
 int _orc_compiler_flag_backup;
+int _orc_compiler_flag_emulate;
 int _orc_compiler_flag_debug;
 
 void
@@ -56,6 +57,7 @@ _orc_compiler_init (void)
   }
 
   _orc_compiler_flag_backup = orc_compiler_flag_check ("backup");
+  _orc_compiler_flag_emulate = orc_compiler_flag_check ("emulate");
   _orc_compiler_flag_debug = orc_compiler_flag_check ("debug");
 }
 
@@ -249,11 +251,11 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
     compiler->valid_regs[i] = 1;
   }
 
-  compiler->target->compiler_init (compiler);
-
   orc_compiler_check_sizes (compiler);
   if (compiler->error) goto error;
 
+  compiler->target->compiler_init (compiler);
+
   orc_compiler_rewrite_insns (compiler);
   if (compiler->error) goto error;
 
@@ -312,6 +314,12 @@ orc_program_compile_full (OrcProgram *program, OrcTarget *target,
     program->orccode->vars[i].value = compiler->vars[i].value;
   }
 
+  if (_orc_compiler_flag_emulate) {
+    program->code_exec = (void *)orc_executor_emulate;
+    compiler->result = ORC_COMPILE_RESULT_UNKNOWN_COMPILE;
+    goto error;
+  }
+
   orc_compiler_assign_rules (compiler);
   if (compiler->error) goto error;
 
@@ -385,6 +393,7 @@ orc_compiler_check_sizes (OrcCompiler *compiler)
 {
   int i;
   int j;
+  int max_size = 1;
 
   for(i=0;i<compiler->n_insns;i++) {
     OrcInstruction *insn = compiler->insns + i;
@@ -407,6 +416,7 @@ orc_compiler_check_sizes (OrcCompiler *compiler)
         compiler->result = ORC_COMPILE_RESULT_UNKNOWN_PARSE;
         return;
       }
+      max_size = MAX(max_size, multiplier * opcode->dest_size[j]);
     }
     for(j=0;j<ORC_STATIC_OPCODE_N_SRC;j++){
       if (opcode->src_size[j] == 0) continue;
@@ -428,6 +438,7 @@ orc_compiler_check_sizes (OrcCompiler *compiler)
         compiler->result = ORC_COMPILE_RESULT_UNKNOWN_PARSE;
         return;
       }
+      max_size = MAX(max_size, multiplier * opcode->src_size[j]);
     }
     if (opcode->flags & ORC_STATIC_OPCODE_SCALAR &&
         opcode->src_size[1] == 0 &&
@@ -439,6 +450,7 @@ orc_compiler_check_sizes (OrcCompiler *compiler)
       return;
     }
   }
+  compiler->max_var_size = max_size;
 }
 
 static OrcStaticOpcode *
@@ -795,6 +807,9 @@ orc_compiler_global_reg_alloc (OrcCompiler *compiler)
           var->ptr_offset = orc_compiler_allocate_register (compiler, FALSE);
           var->aligned_data = orc_compiler_allocate_register (compiler, TRUE);
         }
+        if (var->need_offset_reg) {
+          var->ptr_offset = orc_compiler_allocate_register (compiler, FALSE);
+        }
         break;
       case ORC_VAR_TYPE_DEST:
         var->ptr_register = orc_compiler_allocate_register (compiler, FALSE);
@@ -984,6 +999,13 @@ orc_compiler_load_constant (OrcCompiler *compiler, int reg, int size,
   compiler->target->load_constant (compiler, reg, size, value);
 }
 
+void
+orc_compiler_load_constant_long (OrcCompiler *compiler, int reg,
+    OrcConstant *constant)
+{
+  compiler->target->load_constant_long (compiler, reg, constant);
+}
+
 int
 orc_compiler_get_temp_constant (OrcCompiler *compiler, int size, int value)
 {
@@ -1010,7 +1032,8 @@ orc_compiler_get_constant (OrcCompiler *compiler, int size, int value)
   }
 
   for(i=0;i<compiler->n_constants;i++){
-    if (compiler->constants[i].value == value) {
+    if (compiler->constants[i].is_long == FALSE &&
+        compiler->constants[i].value == value) {
       break;
     }
   }
@@ -1019,6 +1042,7 @@ orc_compiler_get_constant (OrcCompiler *compiler, int size, int value)
     compiler->constants[i].value = value;
     compiler->constants[i].alloc_reg = 0;
     compiler->constants[i].use_count = 0;
+    compiler->constants[i].is_long = FALSE;
   }
 
   compiler->constants[i].use_count++;
@@ -1031,6 +1055,78 @@ orc_compiler_get_constant (OrcCompiler *compiler, int size, int value)
   return tmp;
 }
 
+int
+orc_compiler_get_constant_long (OrcCompiler *compiler,
+    orc_uint32 a, orc_uint32 b, orc_uint32 c, orc_uint32 d)
+{
+  int i;
+  int tmp;
+
+  for(i=0;i<compiler->n_constants;i++){
+    if (compiler->constants[i].is_long == TRUE &&
+        compiler->constants[i].full_value[0] == a &&
+        compiler->constants[i].full_value[1] == b &&
+        compiler->constants[i].full_value[2] == c &&
+        compiler->constants[i].full_value[3] == d) {
+      break;
+    }
+  }
+  if (i == compiler->n_constants) {
+    compiler->n_constants++;
+    compiler->constants[i].full_value[0] = a;
+    compiler->constants[i].full_value[1] = b;
+    compiler->constants[i].full_value[2] = c;
+    compiler->constants[i].full_value[3] = d;
+    compiler->constants[i].is_long = TRUE;
+    compiler->constants[i].alloc_reg = 0;
+    compiler->constants[i].use_count = 0;
+  }
+
+  compiler->constants[i].use_count++;
+
+  if (compiler->constants[i].alloc_reg != 0) {;
+    return compiler->constants[i].alloc_reg;
+  }
+  tmp = orc_compiler_get_temp_reg (compiler);
+  orc_compiler_load_constant_long (compiler, tmp, &compiler->constants[i]);
+  return tmp;
+}
+
+int
+orc_compiler_try_get_constant_long (OrcCompiler *compiler,
+    orc_uint32 a, orc_uint32 b, orc_uint32 c, orc_uint32 d)
+{
+  int i;
+
+  for(i=0;i<compiler->n_constants;i++){
+    if (compiler->constants[i].is_long == TRUE &&
+        compiler->constants[i].full_value[0] == a &&
+        compiler->constants[i].full_value[1] == b &&
+        compiler->constants[i].full_value[2] == c &&
+        compiler->constants[i].full_value[3] == d) {
+      break;
+    }
+  }
+  if (i == compiler->n_constants) {
+    compiler->n_constants++;
+    compiler->constants[i].full_value[0] = a;
+    compiler->constants[i].full_value[1] = b;
+    compiler->constants[i].full_value[2] = c;
+    compiler->constants[i].full_value[3] = d;
+    compiler->constants[i].is_long = TRUE;
+    compiler->constants[i].alloc_reg = 0;
+    compiler->constants[i].use_count = 0;
+  }
+
+  compiler->constants[i].use_count++;
+
+  if (compiler->constants[i].alloc_reg != 0) {;
+    return compiler->constants[i].alloc_reg;
+  }
+  return ORC_REG_INVALID;
+}
+
+
 int
 orc_compiler_get_constant_reg (OrcCompiler *compiler)
 {
