@@ -161,7 +161,7 @@ orc_executor_set_m (OrcExecutor *ex, int m)
 }
 
 static void
-load_constant (void *data, int size, int value)
+load_constant (void *data, int size, orc_uint64 value)
 {
   switch (size) {
     case 1:
@@ -229,6 +229,11 @@ orc_executor_emulate (OrcExecutor *ex)
 
   memset (&opcode_ex, 0, sizeof(opcode_ex));
 
+  if (code == NULL) {
+    ORC_ERROR("attempt to run program that failed to compile");
+    ORC_ASSERT(0);
+  }
+
   if (code->is_2d) {
     m = ORC_EXECUTOR_M(ex);
   } else {
@@ -264,18 +269,28 @@ orc_executor_emulate (OrcExecutor *ex)
       if (var->vartype == ORC_VAR_TYPE_CONST) {
         opcode_ex[j].src_ptrs[k] = tmpspace[insn->src_args[k]];
         /* FIXME hack */
-        load_constant (tmpspace[insn->src_args[k]], 4,
+        load_constant (tmpspace[insn->src_args[k]], 8,
             var->value.i);
       } else if (var->vartype == ORC_VAR_TYPE_PARAM) {
         opcode_ex[j].src_ptrs[k] = tmpspace[insn->src_args[k]];
         /* FIXME hack */
-        load_constant (tmpspace[insn->src_args[k]], 4,
-            ex->params[insn->src_args[k]]);
+        load_constant (tmpspace[insn->src_args[k]], 8,
+            (orc_uint64)ex->params[insn->src_args[k]] |
+            (((orc_uint64)ex->params[insn->src_args[k] +
+             (ORC_VAR_T1 - ORC_VAR_P1)])<<32));
       } else if (var->vartype == ORC_VAR_TYPE_TEMP) {
         opcode_ex[j].src_ptrs[k] = tmpspace[insn->src_args[k]];
       } else if (var->vartype == ORC_VAR_TYPE_SRC) {
+        if (((unsigned long)ex->arrays[insn->src_args[k]]) & (var->size - 1)) {
+          ORC_ERROR("Unaligned array for src%d, program %s",
+              (insn->src_args[k]-ORC_VAR_S1), ex->program->name);
+        }
         opcode_ex[j].src_ptrs[k] = ex->arrays[insn->src_args[k]];
       } else if (var->vartype == ORC_VAR_TYPE_DEST) {
+        if (((unsigned long)ex->arrays[insn->src_args[k]]) & (var->size - 1)) {
+          ORC_ERROR("Unaligned array for dest%d, program %s",
+              (insn->src_args[k]-ORC_VAR_D1), ex->program->name);
+        }
         opcode_ex[j].src_ptrs[k] = ex->arrays[insn->src_args[k]];
       }
     }
@@ -290,6 +305,10 @@ orc_executor_emulate (OrcExecutor *ex)
         opcode_ex[j].dest_ptrs[k] =
           &ex->accumulators[insn->dest_args[k] - ORC_VAR_A1];
       } else if (var->vartype == ORC_VAR_TYPE_DEST) {
+        if (((unsigned long)ex->arrays[insn->dest_args[k]]) & (var->size - 1)) {
+          ORC_ERROR("Unaligned array for dest%d, program %s",
+              (insn->dest_args[k]-ORC_VAR_D1), ex->program->name);
+        }
         opcode_ex[j].dest_ptrs[k] = ex->arrays[insn->dest_args[k]];
       }
     }
@@ -344,7 +363,7 @@ orc_executor_emulate (OrcExecutor *ex)
   }
 
   free (opcode_ex);
-  for(i=0;i<ORC_VAR_T15+1;i++){
+  for(i=0;i<ORC_N_COMPILER_VARIABLES;i++){
     if (tmpspace[i]) free (tmpspace[i]);
   }
 }
