@@ -41,6 +41,7 @@ orc_target_c_get_typedefs (void)
     "typedef uint16_t orc_uint16;\n"
     "typedef uint32_t orc_uint32;\n"
     "typedef uint64_t orc_uint64;\n"
+    "#define ORC_UINT64_C(x) UINT64_C(x)\n"
     "#elif defined(_MSC_VER)\n"
     "typedef signed __int8 orc_int8;\n"
     "typedef signed __int16 orc_int16;\n"
@@ -50,6 +51,7 @@ orc_target_c_get_typedefs (void)
     "typedef unsigned __int16 orc_uint16;\n"
     "typedef unsigned __int32 orc_uint32;\n"
     "typedef unsigned __int64 orc_uint64;\n"
+    "#define ORC_UINT64_C(x) (x##Ui64)\n"
     "#else\n"
     "#include <limits.h>\n"
     "typedef signed char orc_int8;\n"
@@ -61,14 +63,16 @@ orc_target_c_get_typedefs (void)
     "#if INT_MAX == LONG_MAX\n"
     "typedef long long orc_int64;\n"
     "typedef unsigned long long orc_uint64;\n"
+    "#define ORC_UINT64_C(x) (x##ULL)\n"
     "#else\n"
     "typedef long orc_int64;\n"
     "typedef unsigned long orc_uint64;\n"
+    "#define ORC_UINT64_C(x) (x##UL)\n"
     "#endif\n"
     "#endif\n"
     "typedef union { orc_int16 i; orc_int8 x2[2]; } orc_union16;\n"
     "typedef union { orc_int32 i; float f; orc_int16 x2[2]; orc_int8 x4[4]; } orc_union32;\n"
-    "typedef union { orc_int64 i; double f; orc_int32 x2[2]; orc_int16 x4[4]; } orc_union64;\n"
+    "typedef union { orc_int64 i; double f; orc_int32 x2[2]; float x2f[2]; orc_int16 x4[4]; } orc_union64;\n"
     "#endif\n";
 }
 
@@ -101,12 +105,12 @@ orc_target_c_get_asm_preamble (void)
     "#define ORC_CLAMP_UL(x) ORC_CLAMP(x,ORC_UL_MIN,ORC_UL_MAX)\n"
     "#define ORC_SWAP_W(x) ((((x)&0xff)<<8) | (((x)&0xff00)>>8))\n"
     "#define ORC_SWAP_L(x) ((((x)&0xff)<<24) | (((x)&0xff00)<<8) | (((x)&0xff0000)>>8) | (((x)&0xff000000)>>24))\n"
-    "#define ORC_SWAP_Q(x) ((((x)&0xffULL)<<56) | (((x)&0xff00ULL)<<40) | (((x)&0xff0000ULL)<<24) | (((x)&0xff000000ULL)<<8) | (((x)&0xff00000000ULL)>>8) | (((x)&0xff0000000000ULL)>>24) | (((x)&0xff000000000000ULL)>>40) | (((x)&0xff00000000000000ULL)>>56))\n"
+    "#define ORC_SWAP_Q(x) ((((x)&ORC_UINT64_C(0xff))<<56) | (((x)&ORC_UINT64_C(0xff00))<<40) | (((x)&ORC_UINT64_C(0xff0000))<<24) | (((x)&ORC_UINT64_C(0xff000000))<<8) | (((x)&ORC_UINT64_C(0xff00000000))>>8) | (((x)&ORC_UINT64_C(0xff0000000000))>>24) | (((x)&ORC_UINT64_C(0xff000000000000))>>40) | (((x)&ORC_UINT64_C(0xff00000000000000))>>56))\n"
     "#define ORC_PTR_OFFSET(ptr,offset) ((void *)(((unsigned char *)(ptr)) + (offset)))\n"
     "#define ORC_DENORMAL(x) ((x) & ((((x)&0x7f800000) == 0) ? 0xff800000 : 0xffffffff))\n"
     "#define ORC_ISNAN(x) ((((x)&0x7f800000) == 0x7f800000) && (((x)&0x007fffff) != 0))\n"
-    "#define ORC_DENORMAL_DOUBLE(x) ((x) & ((((x)&0x7ff0000000000000ULL) == 0) ? 0xfff0000000000000ULL : 0xffffffffffffffffULL))\n"
-    "#define ORC_ISNAN_DOUBLE(x) ((((x)&0x7ff0000000000000ULL) == 0x7ff0000000000000ULL) && (((x)&0x000fffffffffffffULL) != 0))\n"
+    "#define ORC_DENORMAL_DOUBLE(x) ((x) & ((((x)&ORC_UINT64_C(0x7ff0000000000000)) == 0) ? ORC_UINT64_C(0xfff0000000000000) : ORC_UINT64_C(0xffffffffffffffff)))\n"
+    "#define ORC_ISNAN_DOUBLE(x) ((((x)&ORC_UINT64_C(0x7ff0000000000000)) == ORC_UINT64_C(0x7ff0000000000000)) && (((x)&ORC_UINT64_C(0x000fffffffffffff)) != 0))\n"
     "#if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L\n"
     "#define ORC_RESTRICT restrict\n"
     "#elif defined(__GNUC__) && __GNUC__ >= 4\n"
@@ -430,26 +434,8 @@ c_get_name_int (char *name, OrcCompiler *p, OrcInstruction *insn, int var)
     if (p->target_flags & ORC_TARGET_C_NOEXEC) {
       sprintf(name,"%s", varnames[var]);
     } else if (p->target_flags & ORC_TARGET_C_OPCODE) {
-      switch (p->vars[var].param_type) {
-        case ORC_PARAM_TYPE_INT:
-          sprintf(name,"((orc_union32 *)(ex->src_ptrs[%d]))->i",
-              var - ORC_VAR_P1 + p->program->n_src_vars);
-          break;
-        case ORC_PARAM_TYPE_FLOAT:
-          sprintf(name,"((orc_union32 *)(ex->src_ptrs[%d]))->i",
-              var - ORC_VAR_P1 + p->program->n_src_vars);
-          break;
-        case ORC_PARAM_TYPE_INT64:
-          sprintf(name,"((orc_union64 *)(ex->src_ptrs[%d]))->i",
-              var - ORC_VAR_P1 + p->program->n_src_vars);
-          break;
-        case ORC_PARAM_TYPE_DOUBLE:
-          sprintf(name,"((orc_union64 *)(ex->src_ptrs[%d]))->i",
-              var - ORC_VAR_P1 + p->program->n_src_vars);
-          break;
-        default:
-          ORC_ASSERT(0);
-      }
+      sprintf(name,"((orc_union64 *)(ex->src_ptrs[%d]))->i",
+          var - ORC_VAR_P1 + p->program->n_src_vars);
     } else {
       switch (p->vars[var].param_type) {
         case ORC_PARAM_TYPE_INT:
@@ -499,9 +485,9 @@ static void
 c_get_name_float (char *name, OrcCompiler *p, OrcInstruction *insn, int var)
 {
   if (insn && (insn->flags & ORC_INSTRUCTION_FLAG_X2)) {
-    sprintf(name, "var%d[%d].x2f", var, p->unroll_index);
+    sprintf(name, "var%d.x2f[%d]", var, p->unroll_index);
   } else if (insn && (insn->flags & ORC_INSTRUCTION_FLAG_X4)) {
-    sprintf(name, "var%d[%d].x4f", var, p->unroll_index);
+    sprintf(name, "var%d.x4f[%d]", var, p->unroll_index);
   } else {
     switch (p->vars[var].vartype) {
       case ORC_VAR_TYPE_CONST:
@@ -699,6 +685,8 @@ c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
 #define UNARY_UQ(a,b) UNARY(a,b)
 #define BINARY_BW(a,b) BINARY(a,b)
 #define BINARY_WL(a,b) BINARY(a,b)
+#define BINARY_LQ(a,b) BINARY(a,b)
+#define BINARY_QL(a,b) BINARY(a,b)
 #define BINARY_LW(a,b) BINARY(a,b)
 #define BINARY_WB(a,b) BINARY(a,b)
 #define UNARY_BW(a,b) UNARY(a,b)
@@ -746,6 +734,8 @@ c_rule_ ## name (OrcCompiler *p, void *user, OrcInstruction *insn) \
 #undef UNARY_D
 #undef BINARY_BW
 #undef BINARY_WL
+#undef BINARY_LQ
+#undef BINARY_QL
 #undef BINARY_LW
 #undef BINARY_WB
 #undef UNARY_BW
@@ -770,22 +760,21 @@ c_rule_loadpX (OrcCompiler *p, void *user, OrcInstruction *insn)
   char dest[40];
   int size = ORC_PTR_TO_INT(user);
 
-  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+  if ((p->target_flags & ORC_TARGET_C_NOEXEC) &&
+      (p->vars[insn->src_args[0]].param_type == ORC_PARAM_TYPE_FLOAT))
+    c_get_name_float (dest, p, insn, insn->dest_args[0]);
+  else
+    c_get_name_int (dest, p, insn, insn->dest_args[0]);
 
   if (p->vars[insn->src_args[0]].vartype == ORC_VAR_TYPE_PARAM) {
     if (p->target_flags & ORC_TARGET_C_NOEXEC) {
       ORC_ASM_CODE(p,"    %s = %s;\n", dest, varnames[insn->src_args[0]]);
     } else if (p->target_flags & ORC_TARGET_C_OPCODE) {
-      if (size == 8) {
-        ORC_ASM_CODE(p,"    %s = ((orc_union64 *)(ex->src_ptrs[%d]))->i;\n",
-            dest, insn->src_args[0] - ORC_VAR_P1 + p->program->n_src_vars);
-      } else {
-        ORC_ASM_CODE(p,"    %s = ((orc_union32 *)(ex->src_ptrs[%d]))->i;\n",
-            dest, insn->src_args[0] - ORC_VAR_P1 + p->program->n_src_vars);
-      }
+      ORC_ASM_CODE(p,"    %s = ((orc_union64 *)(ex->src_ptrs[%d]))->i;\n",
+          dest, insn->src_args[0] - ORC_VAR_P1 + p->program->n_src_vars);
     } else {
       if (size == 8) {
-        ORC_ASM_CODE(p,"    %s = (ex->params[%d] & 0xffffffff) | (((orc_uint64)(ex->params[%d + (ORC_VAR_T1 - ORC_VAR_P1)]) << 32);\n",
+        ORC_ASM_CODE(p,"    %s = (ex->params[%d] & 0xffffffff) | ((orc_uint64)(ex->params[%d + (ORC_VAR_T1 - ORC_VAR_P1)]) << 32);\n",
             dest, insn->src_args[0], insn->src_args[0]);
       } else {
         ORC_ASM_CODE(p,"    %s = ex->params[%d];\n", dest,
@@ -799,7 +788,7 @@ c_rule_loadpX (OrcCompiler *p, void *user, OrcInstruction *insn)
           (int)p->vars[insn->src_args[0]].value.i,
           p->vars[insn->src_args[0]].value.f);
     } else {
-      ORC_ASM_CODE(p,"    %s = 0x%08x%08x; /* %gf */\n", dest,
+      ORC_ASM_CODE(p,"    %s = ORC_UINT64_C(0x%08x%08x); /* %gf */\n", dest,
           (orc_uint32)(((orc_uint64)p->vars[insn->src_args[0]].value.i)>>32),
           ((orc_uint32)p->vars[insn->src_args[0]].value.i),
           p->vars[insn->src_args[0]].value.f);
@@ -1189,7 +1178,7 @@ c_rule_convdl (OrcCompiler *p, void *user, OrcInstruction *insn)
   ORC_ASM_CODE(p, "    {\n");
   ORC_ASM_CODE(p,"       int tmp;\n");
   ORC_ASM_CODE(p,"       tmp = %s;\n", src);
-  ORC_ASM_CODE(p,"       if (tmp == 0x80000000 && !(%s&0x8000000000000000ULL)) tmp = 0x7fffffff;\n", src_i);
+  ORC_ASM_CODE(p,"       if (tmp == 0x80000000 && !(%s & ORC_UINT64_C(0x8000000000000000))) tmp = 0x7fffffff;\n", src_i);
   ORC_ASM_CODE(p,"       %s = tmp;\n", dest);
   ORC_ASM_CODE(p, "    }\n");
 }
@@ -1274,6 +1263,30 @@ c_rule_maxd (OrcCompiler *p, void *user, OrcInstruction *insn)
   ORC_ASM_CODE(p,"    }\n");
 }
 
+static void
+c_rule_swapwl (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src[40];
+
+  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+  c_get_name_int (src, p, insn, insn->src_args[0]);
+
+  ORC_ASM_CODE(p,"    %s = ((%s&0x0000ffff) << 16) | ((%s&0xffff0000) >> 16);\n",
+      dest, src, src);
+}
+
+static void
+c_rule_swaplq (OrcCompiler *p, void *user, OrcInstruction *insn)
+{
+  char dest[40], src[40];
+
+  c_get_name_int (dest, p, insn, insn->dest_args[0]);
+  c_get_name_int (src, p, insn, insn->src_args[0]);
+
+  ORC_ASM_CODE(p,"    %s = (ORC_UINT64_C(%s&0x00000000ffffffff) << 32) | (ORC_UINT64_C(%s&0xffffffff00000000) >> 32);\n",
+      dest, src, src);
+}
+
 static OrcTarget c_target = {
   "c",
   FALSE,
@@ -1318,6 +1331,8 @@ orc_c_init (void)
 #define UNARY_D(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define BINARY_BW(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define BINARY_WL(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
+#define BINARY_LQ(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
+#define BINARY_QL(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define BINARY_LW(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define BINARY_WB(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
 #define UNARY_BW(a,b) orc_rule_register (rule_set, #a , c_rule_ ## a, NULL);
@@ -1381,5 +1396,7 @@ orc_c_init (void)
   orc_rule_register (rule_set, "maxf", c_rule_maxf, NULL);
   orc_rule_register (rule_set, "mind", c_rule_mind, NULL);
   orc_rule_register (rule_set, "maxd", c_rule_maxd, NULL);
+  orc_rule_register (rule_set, "swapwl", c_rule_swapwl, NULL);
+  orc_rule_register (rule_set, "swaplq", c_rule_swaplq, NULL);
 }
 
